/*
	tcp/ip dbf server by rb Nov 10/03
	serve up .dbf file data by passing
	on port 6060 a record of the form
	R,dbfname,recno and returning with
	the record of data 

	support write back of a record with
	W,dbfname,recno,data1,data2,data3
	where data1-3 must match the dbf file structure exactly
*/

#include <stdio.h>
#include <unistd.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <strings.h>
#include <string.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <errno.h>
#include <sys/wait.h>
#include <signal.h>
#include "dbc.h"
#include "myfprot.h"

#define SUCCESS 0
#define ERROR	 1

#define	MAXLINE	512
#define	SERV_TCP_PORT	6061

char *datadir = "/u/data/";
char *slfiln = "dbfserver";

static int getout = 0;
extern dBFIL dBFILS[];

void sigexit(int i)
{
	char *s;

	s = "Aborting dbfserver... Stopped by remote user!";
	cerror(s);
	getout = 1;
}

void repln(int mfd, int fl, char *s)
{
	char *ts;

	if((ts = (char *)malloc(528)) != NULL)
		{
		strcpy(ts, s);				/* dont screw with passed str.	*/
		replfld(mfd, ts, fl);
		free(ts);
		}
}

void fetchn(int mfd, int fl, char *s)
{
	dBFIL *dbsptr;
	dBMFIELD *ftmp;

	*s = '\0';						/* set up 's' in case of failure	*/
	dbsptr = &dBFILS[mfd];
	ftmp = &dbsptr->fields[fl];
	strcpy(s, ftmp->fieldnm);
	trim(s);
	strcat(s, "=");
	dfread(dbsptr, s + strlen(s), ftmp);
}

int fldcount(int mfd)
{	
	dBFIL *dbsptr;

	dbsptr = &dBFILS[mfd];
	return dbsptr->nofields;
}

/*
	Pass all uses through here
*/
muse(char *fn)
{
	char gfn[64];

	return(use(catstr(gfn, datadir, fn, NULL), NULL, 0));
}

char *parseany(char *s, char *s1, char c)
{
	char *p;

	p = strchr(s, c);
	if(p != NULL)
	{
		*p = '\0';
		strcpy(s1, s);
		return(p + 1);
	}
	else
	{
		strcpy(s1, s);
		return(s + strlen(s));
	}
}

char *parse(char *s, char *s1)
{
	return parseany(s, s1, 0x9);
}

/*
	data received from machine
	process and return with result
*/
void processData(char *data, char *outp)
{
	int fd, i, cmdrec;
	char *p, *p1, fldnm[16], tmp[4096], fn[4096], s[4096];
	unsigned char errflag;

/* cerror(data); */

	p = parse(data, tmp);
	strcpy(outp, "ERROR");
	switch (*tmp)
	{
		case 'R':
			p = parse(p, fn);
			fd = muse(fn);
			if(fd != -1)
			{
				p = parse(p, tmp);
				cmdrec = atoi(tmp);
				if(cmdrec < 1 || cmdrec > reccount(fd))
				{
					sprintf(s, "%s - read rec# %d - %s", fn, cmdrec, tmp);
					cerror(s);
				}
				else
				{ 
					gorec(fd, cmdrec);
					if (! eof(fd))
					{	
						strcpy(outp, "");
						for(i = 0; i < fldcount(fd); i++)
						{
							fetchn(fd, i, tmp);
							if(i)
								strcat(outp, "\t");
							trim(tmp);
							strcat(outp, tmp);
						}
					}
				}
				closed(fd);
			}
			else
			{
				sprintf(s, "Open ERROR read: %s", fn);
				cerror(s);
			}
			break;
		case 'W':
		case 'A':
			p = parse(p, fn);
			fd = muse(fn);
			if(fd != -1)
			{
				errflag = 0;
				if (*tmp == 'A')
				{
					appendb(fd);
				}
				else
				{
					p = parse(p, tmp);
					cmdrec = atoi(tmp);
					if(cmdrec < 1 || cmdrec > reccount(fd))
					{
						sprintf(s, "%s - write rec# %d - %s", fn, cmdrec, tmp);
						cerror(s);
						errflag = 1;
					}
					else
						gorec(fd, cmdrec);
				}
				if (! errflag)
				{
					while(1)
					/*for(i = 0; i < fldcount(fd); i++)*/
					{
						p = parse(p, tmp);
						if(! *tmp)
							break;
						if(strlen(tmp) < 3)
							break;
						p1 = parseany(tmp, fldnm, '=');
						replace(fd, fldnm, p1);
					}
					strcpy(outp, "DONE");
				}
				closed(fd);
			}
			else
			{
				sprintf(s, "Open ERROR write: %s", fn);
				cerror(s);
			}
			break;
		case 'L':
			p = parse(p, fn);
			fd = muse(fn);
			if(fd != -1)
			{
				p = parse(p, tmp);
				strcpy(fldnm, tmp);
				p = parse(p, tmp);
				cmdrec = atoi(tmp);
				if(ilocate(fd, fldnm, cmdrec))
				{
					strcpy(outp, "");
					for(i = 0; i < fldcount(fd); i++)
					{
						fetchn(fd, i, tmp);
						if(i)
							strcat(outp, "\t");
						trim(tmp);
						strcat(outp, tmp);
					}
				}
				closed(fd);
			}
			break;
	}	
/* cerror(outp); */
}

backgrnd()
{
}

int writen(int fd, char *ptr, int nbytes) 
{
  int nleft, nwritten;
  
  nleft = nbytes;
  while (nleft > 0) 
  {
    nwritten = write(fd, ptr, nleft);
    if (nwritten <= 0)
      return(nwritten);		/* error */
    
    nleft -= nwritten;
    ptr   += nwritten;
  }
  return(nbytes - nleft);
}

int readline(int fd, char *ptr, int maxlen) 
{
  int n, rc;
  char c;

  for (n = 1; n < maxlen; n++) 
  {
    if ( (rc = read(fd, &c, 1)) == 1) 
    {
      *ptr++ = c;
      if (c == '\n')
	    break;
    } 
    else 
      if (rc == 0) 
      {
        if (n == 1)
  	      return(0);	/* EOF, no data read */
/*        else
	      break;		 EOF, some data was read */
      } 
      else
        return(-1);	/* error */
  }
  
  *ptr = 0;
  return(n);
}

/* ------------------------------------------------------------------------ */
/*
 * Read a stream socket one line at a time, and write each line back
 * to the sender.
 *
 * Return when the connection is terminated.
 */

void str_echo(int sockfd) 
{
  int n;
  char line[MAXLINE], outline[4096];

  while(1)
  {
    n = readline(sockfd, line, MAXLINE);
    if (n == 0)
      return;		/* connection terminated */
	else 
      if (n < 0) 
        fprintf(stderr, "str_echo: readline error");
	/* process the incoming command */
	processData(line, outline);
    
	n = strlen(outline);
    if (writen(sockfd, outline, n) != n)
      fprintf(stderr, "str_echo: writen error");
  }
}

static void sig_child(int i)
{
	waitpid((pid_t) -1, NULL, WNOHANG);
	return;
}

void catch_children(void)
{
	struct sigaction handle_sigchild;

	handle_sigchild.sa_handler = &sig_child;
	sigemptyset(&handle_sigchild.sa_mask);

	handle_sigchild.sa_flags = SA_RESTART;

	if(sigaction(SIGCHLD, &handle_sigchild, NULL) != 0)
		perror("catch_children(): sigaction error");
}

/* ------------------------------------------------------------------------ */

int main(int argc, char *argv) 
{
  int sockfd, newsockfd, clilen, childpid;
  struct sockaddr_in cli_addr, serv_addr;

  /*
   * Open a TCP socket (an Internet stream socket).
   */
  
  if ( (sockfd = socket(AF_INET, SOCK_STREAM, 0)) < 0) 
  {
    perror("socket()");
    return(errno);
  }
  
  /*
   * Bind our local address so that the client can send to us.
   */
  
  bzero((char *) &serv_addr, sizeof(serv_addr));
  serv_addr.sin_family      = AF_INET;
  serv_addr.sin_addr.s_addr = htonl(INADDR_ANY);
  serv_addr.sin_port        = htons(SERV_TCP_PORT);
  
  if (bind(sockfd, (struct sockaddr *) &serv_addr, sizeof(serv_addr)) < 0) 
  {
    perror("bind()");
    return(errno);
  }
  
  if (listen(sockfd, 5) < 0) 
  {
    perror("listen()");
    return(errno);
  }
                
  while(1)
  {
    /*
     * Wait for a connection from a client process.
     * This is an example of a concurrent server.
     */
    
    clilen = sizeof(cli_addr);
	/* check for Zombies */
	waitpid((pid_t) -1, NULL, WNOHANG);

    newsockfd = accept(sockfd, (struct sockaddr *) &cli_addr,
		       &clilen);
    if (newsockfd < 0) 
    {
      perror("accept()");
      return(errno);
    }
    
    if ( (childpid = fork()) < 0) 
    {
      perror("fork()");
      return(errno);
    } else 
      if (childpid == 0) 
      {	/* child process */
        close(sockfd);		/* close original socket */
        str_echo(newsockfd);	/* process the request */
        exit(0);
      }

    close(newsockfd);		/* parent process */
  }
}

